#!/usr/bin/python

# This script updates bug descriptions to include PCI
# info from the original reporter's lspci file

# TODO:
#   * (DONE) Implement dryrun throughout
#   * (DONE) Put in system environment printout
#   * For each item, identify how to get it
#   * Remove exception handling (maybe)

from arsenal_lib import *
from lpltk import LaunchpadService

class SystemEnvironment:
    def __init__(self):
        '''Construct the SystemEnvironment object'''
        self.sysenv = [
            { 'name':'chipset',                'value':'' },
            { 'name':'system-architecture',    'value':'' },
            { 'name':'xf86-video-intel',       'value':'' },
            { 'name':'xserver',                'value':'' },
            { 'name':'mesa',                   'value':'' },
            { 'name':'libdrm',                 'value':'' },
            { 'name':'kernel',                 'value':'' },
            { 'name':'linux-distro',           'value':'' },
            { 'name':'machine-model',          'value':'' },
            { 'name':'display-connectors',     'value':'' },
            ]
        
    def load(self, text):
        '''
        Loads object values from a text block, starting
        with the line '[System Environment]' and ending with the
        first line not containing a ':'.
        '''
        regex = re.compile('\s*([^:]+)\s*:\s*(.*)$')

        lines = text.split('\n')
        for l in lines:
            if l == "[System Environment]":
                in_sysenv_section = True
                remainder += "\n"

            elif in_sysenv_section:
                m = regex.match(l)
                if m:
                    key = m.group(1)
                    val = m.group(2)
                    self.set(key, val)
                else:
                    in_sysenv_section = False

            else:
                remainder += l
        
        return remainder

    def set(self, key, val):
        for i in self.sysenv:
            if i['name'] != key:
                continue
            i['value'] = val
            return True
        return False

    def get(self, key):
        for i in self.sysenv:
            if i['name'] == key:
                return i['value']
        return None

    def to_text(self):
        text = "[System Environment]\n"
        for item in self.sysenv:
            text += "%-21s %s\n" % ( item['name']+':', item['value'] )
        return text

def parse_lspci(pci):
    # Parse lspci info in 'pci'
    want_next_subsystem_line = 0
    pci_vga = ''
    for line in pci.split('\n'):
        if want_next_subsystem_line == 1:
            if "Subsystem" in line:
                pci_vga += "    " + line + "\n"
                want_next_subsystem_line = 0
                
        if "00:00.0" in line:
            pci_vga += line + "\n"
            want_next_subsystem_line = 1
        elif "VGA compatible controller" in line:
            pci_vga += line + "\n"
            want_next_subsystem_line = 1
    return pci_vga

parsers = [
    [ 'name':'xorglog', 're':'xorg.*log',   'parser':parse_xorg_log ],
    [ 'name':'lspci',   're':'.*lspci.*',   'parser':parse_lspci ],
    [ 'name':'lshal',   're':'.*lshal.*',   'parser':parse_lshal ],
    [ 'name':'xrandr',  're':'.*xrandr.*',  'parser':parse_xrandr ],
    [ 'name':'system',  're':'system\.txt', 'parser':parse_system ],
    [ 'name':'xdpyinfo', 're':'xdpyinfo.*', 'parser':parse_xdpyinfo ],
    ]

lp = LaunchpadService()
d = lp.load_project("ubuntu")
dryrun = True

if len(sys.argv) < 2:
    sys.stderr.write("Usage: " + sys.argv[0] + " <source-package>\n")
    exit(1)

source_pkgs = sys.argv[1:]

total_updated = 0
for pkg in source_pkgs:
    try:
        lp.reset()
        d = lp.load_project("ubuntu")
        print "Processing ", pkg
        num_updated = 0

        s = d.getSourcePackage(name = pkg)
        for bugtask in s.searchTasks():
            print bugtask.bug.id
            if bugtask.bug_target_display_name != pkg + " (Ubuntu)":
                continue

            bug = ArsenalBug(bugtask.bug, lp.launchpad)
            text = "\n%-8d  %-12s  %s" % (bug.id, bugtask.status, bug.title)
            print text.encode("utf-8")

            sysenv = SystemEnvironment()

            # Parse [System Environment]
            remaining_desc = sysenv.load(bug.description)

            sysenv.set('system-architecture', 'i386')
            sysenv.set('linux-distro', 'Ubuntu')
            sysenv.set('machine-model', 'product board_vendor board_product')

            # TODO:  Parse description(remaining_desc)
            
            
            # Scan attachments for ones with useful system information
            pci = ''
            for a in bug.bug.attachments:
                # if author != bug original reporter, skip it
                author = a.message.owner.name
                if author != bug.bug.owner.name:
                    continue

                hosted_file = a.data
                hosted_file_buffer = hosted_file.open()
                fname = hosted_file.filename

                # If fname is one of interest, get content and parse it
                if parsers.has_key(fname):
                    content = hosted_file_buffer.read()
                
                # TODO:  Parse Xorg.0.log

                # Stop looking at attachments if this one has the PCI info
                if regex_pci.search(content):
                    parse_pci(sysenv, content)
                    exit

            pci_vga = parse_lspci(pci)

            # If there is new info available, print it out
            text = sysenv.to_text()
            if orig_text != text:
                if not dryrun:
                    # TODO:  Update bug description
                    pass
                print text

            # Append pci info to description
            if pci_vga:
                if not dryrun:
                    try:
                        bug.append_description("\n[lspci]\n" + pci_vga)
                        bug.bug.lp_save()
                    except HTTPError, e:
                        print "*** Encountered HTTP error processing ", pkg, " ***"
                        print e.content
                        time.sleep(60)
                        lp.reset()
                        d = lp.load_project("ubuntu")
                print " --> Appended to description:\n", pci_vga
                num_updated += 1
            else:
                print " --> Need PCI info but not available"

        total_updated += num_updated
        print "Number of bugs updated:  ", num_updated
        time.sleep(10)
    except HTTPError, e:
        print "*** Encountered HTTP error ***"
        print e.content
        print "Total updated this time for ", pkg, ":  ", total_updated
        time.sleep(60)
        lp.reset()
        d = lp.load_project("ubuntu")

print "Total updated:  ", total_updated
