#!/usr/bin/python

# This script works on X crash bug reports, moving bugs to
# more appropriate locations if necessary, and extracting
# useful info where possible.

from arsenal_lib import *
import httplib2
from launchpadlib.errors import HTTPError
from lpltk import LaunchpadService

total_count = 0
repackaged  = 0
lp          = LaunchpadService()
d           = lp.load_project("ubuntu")
dryrun      = False
source_packages = ["mesa", "xorg-server"]

drivers = {
    'openchrome': 'xserver-xorg-video-openchrome',
    'intel':      'xserver-xorg-video-intel',
    'nv':         'xserver-xorg-video-nv',
    'nouveau':    'xserver-xorg-video-nouveau',
    'nvidia':     'nvidia-graphics-drivers-180',
    'fglrx':      'fglrx-installer',
    'radeon':     'xserver-xorg-video-ati',
    'radeonhd':   'xserver-xorg-video-radeonhd',
    'savage':     'xserver-xorg-video-savage',
    'sis':        'xserver-xorg-video-sis',
    'r128':       'xserver-xorg-video-r128',
    'mga':        'xserver-xorg-video-mga',
    'psb':        'xserver-xorg-video-psb',
    'vbox':       'virtualbox-ose',
    'vmware':     'xserver-xorg-video-vmware',
    }

title_regexes = [
    {'name':'Nvidia',       're':'(nvidia|GeForce)',        'src':'nvidia-graphics-drivers-180'},
    {'name':'Radeonhd',     're':'radeonhd',                'src':'xserver-xorg-video-radeonhd'},
    {'name':'Fglrx',        're':'fglrx',                   'src':'fglrx-installer'},
    {'name':'VirtualBox',   're':'vboxvideo',               'src':'virtualbox-ose'},
    {'name':'VMware',       're':'vmware',                  'src':'xserver-xorg-video-vmware'},
]
for title_regex in title_regexes:
    title_regex['rc'] = re.compile(title_regex['re'], re.IGNORECASE)

pci_devices = [
    { 'name':'ati',         're':'ATI Technologies '},
    { 'name':'nvidia',      're':'(nvidia|GeForce)'},
    { 'name':'intel',       're':'Intel Corporation'},
    { 'name':'openchrome',  're':'VIA Technologies'},
    { 'name':'vmware',      're':'VMware SVGA'},
    ]
for device in pci_devices:
    device['rc'] = re.compile(device['re'], re.IGNORECASE)

regex_nonfree_kernel_fglrx  = re.compile('(NonfreeKernelModules|SourcePackage): *fglrx')
regex_nonfree_kernel_nvidia = re.compile('(NonfreeKernelModules|SourcePackage): *nvidia')
regex_virtualbox = re.compile('dmi\.bios\.version: VirtualBox')
regex_vmware = re.compile('dmi\.product\.name: VMware Virtual Platform')

def transitionToTarget(bugtask, package_name):
    if package_name != None:
        package = d.getSourcePackage(name=package_name)
        if not package or not package.display_name:
            print "Error:  Could not find source package ",package_name
            return False

        for task in bug.bug.bug_tasks:
            if not task.target.display_name:
                print "Error: No display_name for target"
                return False
            if task.target.display_name == package.display_name:
                # Already has a task against this target
                print "--> Already has a task against this target"
                return True

        if not dryrun:
            bugtask.transitionToTarget(target = package)
    return True

def get_pci_device(desc):
    regex_vga = re.compile('VGA compatible controller: (.*)$', re.IGNORECASE)

    lines = regex_vga.findall(desc)
    if len(lines) > 0:
        for l in lines:
            for device in pci_devices:
                if device['rc'].search(l):
                    return device['name']

    return None

def get_driver_from_xorglog(bug):
    regex_xorglog = re.compile("xorg.*log", re.IGNORECASE)
    regex_driver = re.compile("\(II\) Loading /usr/lib/xorg/modules/drivers/+(.*)_drv\.so")
    drivers = []
    for a in bug.attachments:
        if not regex_xorglog.search(a.title):
            continue

        # if author != bug original reporter, skip it
        if a.message.owner.name != bug.owner.name:
            continue

        hosted_file = a.data
        hosted_file_buffer = hosted_file.open()
        content = hosted_file_buffer.read()

        # Parse data, look for driver
        found_drivers = regex_driver.findall(content)

        video_driver = None
        for d in found_drivers:
            if video_driver and video_driver != d:
                # Ambiguous; multiple video drivers are loaded.  Needs human review
                return None
            video_driver = d

        if video_driver:
            drivers.append(video_driver)

    last_driver = None
    for d in drivers:
        if last_driver != None and d != last_driver:
            # Ambiguous; some logs show one driver, others another.  Needs human review
            return None
        last_driver = d

    return last_driver


for source_package in source_packages:
    print "Processing %s" %(source_package)
    s = d.getSourcePackage(name = source_package)
    for bugtask in s.searchTasks(assignee=None):
        bug = ArsenalBug(bugtask.bug, lp.launchpad)
        if not (bug.has_tag("crash") or bug.has_tag("apport-crash") or bug.has_tag("crash-nullptr")):
            continue
        if bug.has_tag("omit"):
            continue
    
        title = bug.title.decode("utf-8")
        desc = bug.description.decode("utf-8")
        pci_device = get_pci_device(desc)
        text = "%-7s %-11s %s" % (bug.id, bugtask.status, title)
    
        # Bugs with a significant status should not be processed
        if bugtask.status == "In Progress" or bugtask.status == "Fix Committed" or bugtask.importance == "Wishlist":
            continue

        # Bugs that appear to have a non-default priority should not be moved
        if not (bugtask.importance == "Medium" or bugtask.importance == "Undecided"):
            continue

        total_count += 1

        title_match = False
        for title_regex in title_regexes:
            if title_regex['rc'].search(title):
                print "%-12s -->  " % (title_regex['name']), text.encode("utf-8")
                transitionToTarget(bugtask, title_regex['src'])
                repackaged += 1
                title_match = True
                break
        if title_match:
            continue

        if regex_nonfree_kernel_fglrx.search(desc):
            print "%-12s -->  " % ("fglrx"), text.encode("utf-8")
            transitionToTarget(bugtask, "fglrx-installer")
            repackaged += 1
            continue

        if regex_nonfree_kernel_nvidia.search(desc):
            print "%-12s -->  " % ("nvidia"), text.encode("utf-8")
            transitionToTarget(bugtask, "nvidia-graphics-drivers")
            repackaged += 1
            continue

        if regex_virtualbox.search(desc):
            print "%-12s -->  " % ("vbox"), text.encode("utf-8")
            transitionToTarget(bugtask, "virtualbox-ose")
            repackaged += 1
            continue

        if regex_vmware.search(desc):
            print "%-12s -->  " % ("vmware"), text.encode("utf-8")
            transitionToTarget(bugtask, "xserver-xorg-video-vmware")
            repackaged += 1
            continue

        try:
            driver = get_driver_from_xorglog(bugtask.bug)

            if driver == 'nvidia':
                print "%-12s -->  " % ("nvidia"), text.encode("utf-8")
                transitionToTarget(bugtask, "nvidia-graphics-drivers")
                repackaged += 1
                continue
            elif driver == 'fglrx':
                print "%-12s -->  " % ("fglrx"), text.encode("utf-8")
                transitionToTarget(bugtask, "fglrx-installer")
                repackaged += 1
                continue
            elif driver == 'vbox':
                print "%-12s -->  " % ("vbox"), text.encode("utf-8")
                transitionToTarget(bugtask, "virtualbox-ose")
                repackaged += 1
                continue
            
        except:
            print "*** Exception:  Couldn't get driver from Xorg.0.log; ignoring ***"

        #print " ???              ", text.encode("utf-8")

print total_count," total bugs, ",repackaged," re-packaged"

