#!/usr/bin/python

# This script is for automatic processing of replies to bugs marked
# Incomplete, to filter out ones where requested info has been provided
# from ones that got incorrectly set to with-response via a
# non-original-reporter's reply.

from arsenal_lib import *
from lpltk import *

if len(sys.argv) < 2:
    sys.stderr.write("Usage: %s <source-package>\n" %(sys.argv[0]))
    exit(1)

opt_dryrun  = 0
total_count = 0
source_pkgs = sys.argv[1:]
lp          = LaunchpadService()
d           = lp.load_project("ubuntu")

total_count = 0

regex_xorglog          = re.compile('xorg.*log',           re.IGNORECASE)
regex_conf             = re.compile('conf',                re.IGNORECASE)
regex_backtrace        = re.compile('ktrace',              re.IGNORECASE)
regex_lspci            = re.compile('lspci',               re.IGNORECASE)
regex_master           = re.compile("MASTER:")
regex_fixed            = re.compile("(fixed now|works now|can be closed)")


if opt_dryrun:
    print "dryrun - no actual changes will be made"

try:
    for source_pkg in source_pkgs:
        count=0
        s = d.getSourcePackage(name = source_pkg)
        for bugtask in s.searchTasks(status="Incomplete (with response)"):
            bug = ArsenalBug(bugtask.bug, lp.launchpad)
            if bug.ubuntu_omittable():
                continue
            
            original_reporter = bug.bug.owner
            if not original_reporter:
                continue
            
            # Skip bugs marked "MASTER:"
            if regex_master.search(bug.title):
                continue

            try:
                text = "%d  %s  %s" %(bug.id, bugtask.status, bug.title)
            except UnicodeDecodeError:
                text = "%d  %s  %s" %(bug.id, bugtask.status, "*** Unicode Exception ***")

            # TODO:  Some of this attachment handling logic should be done in arsenal_lib
            #        Probably needs a new ArsenalAttachment wrapper class
            # TODO:  It would be nice to define pre-conditions in a 'rules' file, that
            #        arsenal could then apply to do triaging more programmatically
            has_xorglog = False
            has_lspci = False
            has_other_files = False
            has_nonowner_info = False
            for a in bug.bug.attachments:
                if not a.message or not a.message.owner:
                    continue
                try:
                    author = a.message.owner.name
                    hosted_file = a.data
                    hosted_file_buffer = hosted_file.open()
                    t = hosted_file_buffer.filename

                    if author == original_reporter.name:
                        if regex_xorglog.search(t):
                            has_xorglog = True
                        elif regex_lspci.search(t):
                            has_lspci = True
                        elif a.message.date_created and (a.message.date_created > bugtask.date_left_new):
                            has_other_files = True
                    elif regex_xorglog.search(t) or regex_lspci.search(t):
                        has_nonowner_info = True

                except:
                    print text
                    print "*** Exception ***  Failed to get contents of attachment"

            # If has tag needs-*, see if those are attached now
            # If so, remove tag.  If no more needs-* tags, set to Confirmed
            got_xorglog = False
            got_lspci_vvnn = False
            if bug.has_tag('needs-xorglog') and has_xorglog == 1:
                if not opt_dryrun:
                    bug.remove_tag('needs-xorglog')
                print text
                print " --> removed needs-xorglog"
                got_xorglog = True

            if bug.has_tag('needs-lspci-vvnn') and has_lspci == 1:
                if not opt_dryrun:
                    bug.remove_tag('needs-lspci-vvnn')
                print text
                print " --> removed needs-lspci-vvnn"
                got_lspci_vvnn = True

            # If all conditions met, set to Confirmed
            if (got_xorglog or got_lspci_vvnn) and \
               (has_xorglog and has_lspci):
                try:
                    if not opt_dryrun:
                        bugtask.status = "Confirmed"
                        bugtask.lp_save()
                    print text
                    print " ---> Confirmed"
                    count += 1
                    total_count += 1
                except:
                    print text
                    print " ERROR:  Failed to update bug to Confirmed ", bug.id
                    raise

            elif has_other_files:
                # Bug reporter supplied some files.  Needs manual review
                pass

            else:
                # Skip bugs where we don't know when they entered Incomplete status
                if not bugtask.date_left_new:
                    print text
                    print " --> Unknown when became incomplete"
                    continue

                # Bugs with lots of comments are probably stuck in a me-too storm
                # In any case, it'll take too long to process them
                if len(bug.bug.messages)>50:
                    print text
                    print " --> Excessive comment quantity(%d) - needs manual review" %(len(bug.bug.messages))
                    continue

                new_status = None
                reason     = None
                message    = None
                has_owner_reply = None

                # Review comments since bug was set to Incomplete
                for m in bug.bug.messages:
                    # Skip messages from before when bug was set to incomplete
                    if (not m.date_created) or (m.date_created < bugtask.date_left_new):
                        continue

                    if m.owner.name == bug.owner.name:
                        if regex_fixed.search(m.content):
                            new_status = "Fix Released"
                            reason = "reported-as-fixed"
                            message = "[From reporter's comment it sounds like this is now fixed.  If not, please reopen with clarification.]"
                        elif got_xorglog and not got_lspci_vvnn:
                            new_status = "Incomplete"
                            reason = "needs-lspci"
                            message = "Thanks for the Xorg log.  We also need the lspci -vnvn output attached to this bug as a file"
                        elif got_lspci_vvnn and not got_xorglog:
                            new_status = "Incomplete"
                            reason = "needs-xorglog"
                            message = "Thanks for the lspci info.  We also need the Xorg.0.log attached to this bug report as a file (not pasted in)."

                        # Owner has commented but it's unclear what the deal is.  Don't change bug.
                        elif not has_owner_reply:
                                reason = "has-owner-reply"
                                new_status = None
                                message = None
                        has_owner_reply = True

                    # If owner is in team ubuntu-x-swat or ubuntu bug control or bugsquad
                    elif is_triager(m.owner):
                        # If all non-original-reporter repliers are bug triagers, reset to Incomplete
                        if not reason:
                            new_status = "Incomplete"
                            reason = "triager-comment"

                    elif not has_owner_reply:
                        # If no one replying to the bug was the original reporter, reset to Incomplete
                        if has_nonowner_info and m.owner.karma < 1000:
                            if not reason:
                                new_status = "Incomplete"
                                reason = "nonowner-supplied-info"
                                message = """
@%s - thanks for providing that info, but we would like to see it from the original reporter.  We may want to send this bug report upstream, but upstream has a hard and fast rule that logs from someone other than the original reporter are not acceptable.  The reason for that is that quite often with X.org bug reports, two people may *think* they have the same issue based on commonality of symptoms and/or hardware, but in truth they have separate issues and mixing the report and the files can cause the developers some confusion.

If this bug report seems not to be getting attention, I would encourage you to file a new bug report with this data, and simply reference this bug report in your bug description.
""" %(m.owner.name)
                        elif m.owner.karma < 100:
                            if not reason:
                                new_status = "Incomplete"
                                reason = "drive-by-commenter"
                                message = "[Resetting to incomplete since we need a response from the original reporter on this]."

                if new_status:
                    print " --> %s (%s)" %(new_status, reason)
                    if not opt_dryrun:
                        if message:
                            if bug.append_comment(message):
                                if new_status == "Incomplete":
                                    bugtask.status = "New"
                                    bugtask.lp_save()
                                bugtask.status = new_status
                                bugtask.lp_save()
                                count += 1
                                total_count += 1
                        else:
                            if new_status == "Incomplete":
                                bugtask.status = "New"
                                bugtask.lp_save()
                            bugtask.status = new_status
                            bugtask.lp_save()
                            count += 1
                            total_count += 1

        if count > 0:
            print count, " bugs updated for ", source_pkg

except:
    if total_count > 0:
        print total_count, " bugs updated"
    raise

if total_count > 0:
    print total_count, " bugs updated"
