#!/usr/bin/python

# The purpose of this script is to conduct the initial review of
# incoming bugs and attempt to get them into a better shape for later
# scripts to handle.
#
# We want to categorize the bugs as best we can, and request appropriate
# information so later scripts can conduct their analysis.  We also want
# to kick out bug reports that aren't valid for handling by the X team.

from arsenal_lib    import *
from lpltk          import *
from getopt         import getopt, GetoptError

# Class to abstract some of the functions of crafting a good reply
# TODO:
#   * Reduce need for indicating automated message
#   * Make the message_* stuff be registered into this class somehow
#   * Vary the message used depending on quantity of items requested
#   * Control whether to specify it's an automated message based on length of reply
#   * Divide up phrasings more cleverly, esp. responses to stuff already included
#   * Include directions for if their request is a wishlist, packaging, or workflow bug
#   * Ignore bugs starting with "Please..."
#   * Check for 'patch' in subject
#   * If a zip or gz file was uploaded in the last 3 messages, auto append request not to do that, except for failsafe X tarballs
#   * If resolution mentioned, request xrandr output
class ArsenalReply:
    automated = """
    [This is an automated message.  Apologies if it has reached you inappropriately; please just reply to this message indicating so.]
    """
        
    messages = {
        'all' : """
        Please run the command 'apport-collect BUGNUMBER', which will attach several files we need for debugging.
        """,
        'img_only' : """
        Thanks for including an image to demonstrate the issue.  Also, please run the command 'apport-collect BUGNUMBER' which will attach additional files we need.
        """,
        'some' : """
        Thanks for including the attached files.  Could you also include your /var/log/Xorg.0.log (or Xorg.0.log.old) from after reproducing the issue?
        """,
        'backtrace' : """
        Could you please collect a full backtrace from when this crash occurs?  Directions on collecting full backtraces are at https://wiki.ubuntu.com/X/Backtracing
        """,
        'also_xorglog' : """
        Also, could you please attach your /var/log/Xorg.0.log from after reproducing the issue?
        """,
        'also_lspci_vvnn' : """
        Please attach the output of `lspci -vvnn` too.
        """,
        'xkeyboard_config' : """
        Please attach the output of 'setxkbmap -print' and 'xkbcomp :0 -'.
        """,
        'old_bug' : """
        Thank you for taking the time to report this bug and helping to make Ubuntu better. You reported this bug a while ago and there hasn't been any activity in it recently. We were wondering is this still an issue for you? Can you try with the latest development release of Ubuntu?  (ISO CD images are available from http://cdimage.ubuntu.com/releases/)
        
        If it remains an issue, could you also attach a new /var/log/Xorg.0.log?
        Thanks in advance.
        """
        }
        
    def __init__(self, bug):
        self.bug = bug
        self.message = ""

    def set_message(self, message, source_package='', bugid=''):
        if (source_package != '') and (bugid != ''):
            self.message = self.messages[message] % (source_package, bugid)
        elif (bugid != ''):
            self.message = self.messages[message] % (bugid)
        else:
            self.message = self.messages[message]

    def append_message(self, message):
        self.message = self.message + message

    def get_message(self):
        return self.message

    def make_reply(self):
        result = ''
        if self.message != "":
            greeting = "Hi " + self.bug.owner_firstname + ",\n\n"
            result = greeting + self.message + self.automated
        return result

class ProcessNewBugsApp:
    opt_dryrun      = False
    opt_source_pkgs = None
    opt_bug_limit   = -1
    opt_verbose     = False

    def __init__(self):
        # Table of regexes for different files
        self.file_regexes = [
            {'name':'system_env', 're':'system', },
        
            {'name':'image',      're':'(png|jpg)', },
            {'name':'xorglog',    're':'xorg.*log', },
            {'name':'gdmlog',     're':':.*log', },
            {'name':'xorgconf',   're':'xorg.*conf', },
            {'name':'xrandr',     're':'xrandr', },
            {'name':'glxinfo',    're':'glxinfo', },
            {'name':'setxkbmap',  're':'setxkbmap', },
            {'name':'monitors',   're':'monitors\.xml', },
        
            {'name':'gpudump',    're':'intelgpudump.txt', },
        
            {'name':'strace',     're':'strace', },
            {'name':'backtrace',  're':'ktrace', },
            {'name':'dri_debug',  're':'dri_debug.*tgz', },
            {'name':'vmstat',     're':'vmstat', },
            {'name':'coredump',   're':'coredump', },
        
            {'name':'deps',       're':'Dependencies\.txt', },
            {'name':'ProcMaps',   're':'ProcMaps\.txt', },
            {'name':'ProcStatus', 're':'ProcStatus\.txt', },
            {'name':'disassembly','re':'Disassembly\.txt', },
            {'name':'registers',  're':'registers', },
            {'name':'dpkg',       're':'dpkg\.log', },
            {'name':'hal',        're':'HalComputerInfo\.txt', },
            {'name':'cpuinfo',    're':'ProcCpuinfo\.txt', },
            {'name':'interrupts', 're':'ProcInterrupts\.txt', },
            {'name':'modules',    're':'ProcModules\.txt', },
            {'name':'dmesg',      're':'dmesg\.txt', },
        
            {'name':'lshw',       're':'lshw', },
            {'name':'lspci',      're':'lspci', },
            {'name':'lsusb',      're':'lsusb', },
            {'name':'lshal',      're':'lshal', },
            {'name':'lsmod',      're':'lsmod', },
        ]
        for file_regex in self.file_regexes:
            file_regex['rc'] = re.compile(file_regex['re'], re.IGNORECASE)
        
        # These correspond to standard ubuntu-x tags
        # TODO:  Figure out how to handle these:
        #    * 'rotate'
        #    * high memory
        #    * font size/dpi
        #    * hotkey issues, ' key '
        #    * 'random' in title or description
        #    * 'failed to install'
        #    * not using .* driver
        #    * 'does not boot'
        #    * 'random'
        #    * kde/kubuntu/konqueror/kwin
        self.symptom_regexes = [
            {'name':'error-exit',    're':'(exits with|server terminated|kills session)', },
            {'name':'error-exit',    're':'(forces a logout)', },
            {'name':'no-screens',    're':'(no screens found|screen not detected)', },
            {'name':'crash',         're':'(crash|gdm_slave_xioerror_handler)', },
            {'name':'freeze',        're':'(freeze|freezing| hangs* |locks* *up|system unresponsive)', },
            {'name':'corruption',    're':'(corrupt|rubbish|garbage|garbled|display bugs|colored noise)', },
            {'name':'tearing',       're':'(tearing)', },
            {'name':'resolution',    're':'(resolution| mode |frequency too high)', },
            {'name':'performance',   're':'(performance|sluggish|jerky|slow at redraw)', },
            {'name':'flicker',       're':'(flicker|blinking|blinks|flashes periodically)', },
            {'name':'dual-head',     're':'(dual|ext|external)[ -]*(head|monitor)', },
            {'name':'backlight',     're':'(backlight|brightness)', },
            {'name':'videoplayback', 're':'( Xv |mplayer|xine|totem)', },
            {'name':'ghost-monitor', 're':'(false|non-existant) (monitor|vga )', },
            {'name':'3d',            're':'(3D |OpenGL| GL |Blender)', },
            {'name':'compiz',        're':'( compiz|desktop effects)', },
            {'name':'resume',        're':'(suspend|resume|hibernate|wake up)', },
            {'name':'cropped',       're':'(cropped)', },
            {'name':'font-size',     're':'(text .* too large|huge font)', },
            {'name':'black-screen',  're':'(black|blank) (tv |video|screen)', },
            {'name':'black-screen',  're':'screen (goes|stays) (black|blank)', },
            {'name':'black-screen',  're':'no X.* display', },
            {'name':'vt-switch',     're':'(VT switch|ctrl-alt-f|tty switch|switch .*to console)', },
            {'name':'tv-out',        're':'(TV[ -]*Out|S[ -]*Video| HDMI )', },
            {'name':'high-cpu',      're':'(high \%*cpu|cpu \d+\%|keeps processor.*busy|high load)', },
            {'name':'mouse-pointer', 're':'(mouse pointer)', },
            ]
        for symptom_regex in self.symptom_regexes:
                symptom_regex['rc'] = re.compile(symptom_regex['re'], re.IGNORECASE)
        
        # TODO:  Generate this automatically (or is it even necessary to pre-seed it?)
        self.has_file = {
            'image':     0,
            'xorglog':   0,
            'gdmlog':    0,
            'conf':      0,
            'gpudump':   0,
            'strace':    0,
            'backtrace': 0,
            'lspci':     0,
            'deps':      0,
            'unknown':   0,
            'dmesg':     0,
        }

        self.regex_subj_app_crash = re.compile('(\w+) crashed with SIGSEGV')
        self.regex_subj_crash     = re.compile('(crash|SIGSEGV)',     re.IGNORECASE)
        self.regex_subj_workflow  = re.compile('(Sync|Merge|Please sync|Please merge|debian/control|depends|requires|translation|\.pot)')

        self.regex_desc_crash     = re.compile('ProblemType: Crash')
        self.regex_desc_xorglog   = re.compile('X.Org X Server')
        

    # cmdline_usage
    #
    # Prints out the help text which explains the command line options.
    #
    def cmdline_usage(self):
        sys.stdout.write("""
    Usage:
        %s [--dryrun] [--limit] <source-package>

    Options:
        --help           Prints this text.

        --dryrun         Don't commit any changes to launchpad, just report what
                         would be done.

        --limit=<n>      Only process <n> number of bugs before exiting.

        --verbose        Give some feedback of what is happening while the script is
                         running.
""" % (sys.argv[0]))

    # cmdline_process
    #
    # As you can probably tell from the name, this method is responsible
    # for calling the getopt function to process the command line. All
    # parameters are processed into class variables for use by other
    # methods.
    #
    def cmdline_process(self):
        result = True
        try:
            optsShort = ''
            optsLong  = ['help', 'dryrun', 'limit=', 'verbose']
            opts, args = getopt(sys.argv[1:], optsShort, optsLong)

            for opt, val in opts:
                if (opt == '--help'):
                    result = False
                    break

                elif (opt == '--dryrun'):
                    self.opt_dryrun = True

                elif (opt == '--limit'):
                    self.opt_bug_limit = int(val)

                elif (opt == '--verbose'):
                    self.opt_verbose = True

            if result:
                if (len(args) != 0):
                    self.opt_source_pkgs = args
                else:
                    print("\n   Error: At least one source package must be specified.")
                    result = False

        except GetoptError, error:
            print(error)
            result = False

        return result

    # skip_check
    #
    # Contains all the special rules which identify bugs that we want to 'skip'
    # and not process.
    #
    def skip_check(self, bugtask, bug):
        result = False

        while True:
            # Skip bugs we don't want to handle automatically
            if bug.ubuntu_omittable():
                break

            # Skip wishlist bugs
            if bugtask.importance == "Wishlist":
                break

            # Skip if bug is assigned to someone
            if bugtask.assignee:
                break

            # Don't try to handle bugs that have been automatically processed before
            if bug.has_text_match("This is an automated message"):
                break

            # TODO: Skip bugs that look like sync requests
            # Subject =~ /Debian/ or has ubuntu-main-sponsors subscribed

            # If bug has more than 3 open ubuntu bugtasks, skip it
            num_tasks = 0
            for task in bug.bug.bug_tasks:
                if not task.is_complete:
                    num_tasks += 1
            if num_tasks > 3:
                #print "  --> Too many open bug tasks.  Needs manual review"
                break

            result = True
            break
        
        return False

    # action_handler
    #
    def action_handler(self, bug, bugtask):
        # TODO:  Distinguish between rules that don't apply for certain kinds of
        #        packages, like not asking for lspci for input devices, or not
        #        bothering to ask for crash info for proprietary drivers, or
        #        so on.
        actions = [
            {
                'files':       ['backtrace', 'xorglog', 'lspci'],
                'no-files':    [],
                'symptoms':    [],
                'no-symptoms': [],
                'reply':       '',
                'status':      'Confirmed',
            },
            {
                'files':       ['xorglog', 'lspci'],
                'no-files':    ['crash'],
                'symptoms':    [],
                'no-symptoms': [],
                'reply':       '',
                'status':      'Confirmed',
            },
            {
                'files':       ['gpudump'],
                'no-files':    [],
                'symptoms':    [],
                'no-symptoms': [],
                'reply':       '',
                'status':      'Confirmed',
            },
        ]
        
        for action in actions:
            do_action = True
            for f in actions['files']:
                if not self.has_file[f]:
                    do_action = False
                    break
            for nf in actions['no-files']:
                if self.has_file[nf]:
                    do_action = False
                    break
            for s in actions['symptoms']:
                if not has_symptom[s]:
                    do_action = False
                    break
            for ns in actions['non-symptoms']:
                if has_symptom[ns]:
                    do_action = False
                    break

            if do_action:
                if action['status']:
                    print bug.id, "  ", bugtask.status, "  ", bug.title
                    print " --> %s (has %s)" % (action['status'], 'needed files') # TODO: list files
                if action['reply']:
                    print bug.id, "  ", bugtask.status, "  ", bug.title
                    print "Would give reply here"  # TODO

    # add_tag
    #
    # One place to append tags to bugs so the dryrun flag gets checked and we produce
    # verbose output if we want.
    #
    def add_tag(self, bug, tag):
        if not self.opt_dryrun:
            bug.append_tags(tag)

        if self.opt_verbose:
            print("  --> Tagged %s" % tag)

    # set_status
    #
    # One place to change a bugs status and produce verbose output if requested.
    #
    def set_status(self, new_status, bug, bugtask):
        if not self.opt_dryrun:
            bugtask.transitionToStatus(status = new_status)

        if self.opt_verbose:
            print("  --> Status change: %s" % (new_status))

    # detect_attachments
    #
    # We build a list of the attachments that have been added to a bug and then use that
    # to determine if we need to ask the user to add specific files.
    #
    def detect_attachments(self, bug):
        if self.regex_desc_xorglog.search(bug.description):
            self.has_file['xorglog'] = 1

        if len(bug.bug.attachments) != 0:
            # Analyze what sorts of files are attached
            for a in bug.bug.attachments:
                try:
                    hosted_file = a.data
                    # The attachment content is downloaded in full during the call to
                    # open(), so this can be slow.
                    hosted_file_buffer = hosted_file.open()
                    t = hosted_file_buffer.filename
                    #content = hosted_file_buffer.read()

                    is_unknown = True
                    for file_regex in self.file_regexes:
                        if not file_regex['rc'].search(t):
                            continue
                    
                        # Flag the hit
                        dbg("  %s [%s]" % (t, file_regex['name']))
                        self.has_file[file_regex['name']] = 1
                        is_unknown = False
                        break

                    if is_unknown:
                        self.has_file['unknown'] = 1
                    # TODO:  Search bug comments and Xorg.0.log's for lspci info, and if not, ask for it
                    # TODO:  For intel, I need dmidecode

                except KeyboardInterrupt:
                    raise

                except:
                    print "*** Exception *** encountered while checking attachment contents"
        

    def main(self):
        if (not self.cmdline_process()):
            self.cmdline_usage()
            return

        # TODO: Only print these for debugging
        #print("Current cmdline parameters:")
        #print("         dryrun: %s" % (self.opt_dryrun))
        #print("    source pkgs: %s" % (self.opt_source_pkgs))
        #print("          limit: %d" % (self.opt_bug_limit))   

        if self.opt_dryrun:
            print('')
            print("  ** Note: This is a 'dry run', no bugs will actually be modified!")
            print('')

        # Currently this script is suffering bug #336866 when trying to
        # apply tags
        
        total_count = 0
        lp          = LaunchpadService()
        d           = lp.load_project("ubuntu")
        new_status  = "Incomplete"
        
        #arsenal.Debug(1)
        
        symptoms = [
            {'name':'freeze',     'source':'xserver-xorg-video-intel', 'needs': ['xorglog', 'lspci', 'dri_debug', 'gpudump'], },
            {'name':'crash',      'source':'any',                      'needs': ['xorglog', 'lspci', 'backtrace'], },
            {'name':'generic',    'source':'any',                      'needs': ['xorglog', 'lspci'], },
            ]
        
        
        # TODO:  These should merge with the symptom/tagging stuff
        
        try:
            examined_count = 0
            for source_pkg in self.opt_source_pkgs:
                updated_count=0
                if self.opt_dryrun:
                    print("\nProcessing %s" % (source_pkg))

                s = d.getSourcePackage(name = source_pkg)
                for bugtask in s.searchTasks(status="New"):
                    if (self.opt_bug_limit != -1) and (examined_count >= self.opt_bug_limit):
                        print("Hit max number of bugs to process, exiting.")
                        break
                    examined_count += 1
 
                    bug = ArsenalBug(bugtask.bug, lp.launchpad)
                    bug_title = bug.title
                    bug_description = bug.description

                    if (self.opt_verbose):
                        print("%s  %s" % (bugtask.status, str(bug.title)))
                        print("  https://bugs.launchpad.net/ubuntu/+bug/%d" % (bug.id))

                    # Should we just skip this one?
                    #
                    if (self.skip_check(bugtask, bug)):
                        if (self.opt_verbose):
                            print("    Skipped")
                        continue

                    # TODO:  Move message construction logic into ArsenalReply
                    if self.regex_subj_workflow.search(bug.title):
                        self.add_tag(bug, "workflow")
                        continue
        
                    # TODO:  Copy tags from https://wiki.ubuntu.com/X/Tagging
                    has_symptom = {
                        'crash':     0,
                    }

                    # Analyze title and description for common issues, then set tags accordingly
                    # TODO:  If no symptoms found in title, perhaps try description??
                    if source_pkg == 'xserver-xorg-video-intel' or source_pkg == 'xserver-xorg-video-ati':
                        for symptom in self.symptom_regexes:
                            # TODO:  Test this with dryrun set first
                            continue
        
                            if symptom['name'] in bug.bug.tags:
                                has_symptom[symptom['name']] = 1
                                print("  --> Already tagged %s" % (symptom['name']))
        
                            elif symptom['rc'].search(title):
                                self.add_tag(bug, symptom['name'])
                                tagged += 1
        
                    reply = ArsenalReply(bug)
                    rationale = None

                    # TODO:  Also check if a backtrace is shown -> 'crash'
                    # TODO:  This prints out before the bug.id stuff
                    if self.regex_subj_crash.search(bug.title) or \
                            self.regex_desc_crash.search(bug.description):
                        rationale = " + crash"
                        has_symptom['crash'] = 1
                        self.add_tag(bug, "crash")
        
                    self.detect_attachments(bug)
        
                    if not self.has_file['xorglog']:
                        rationale = " + Needs xorglog"
                        reply.set_message('all')
                        self.add_tag(bug, "needs-xorglog")
                        self.add_tag(bug, "needs-lspci-vvnn")
                        print("*******************************    Xorglog") # FIXME bjf

                    if not self.has_file['xorglog'] and not self.has_file['unknown']:
                        self.add_tag(bug, "needs-xorglog")
                        if self.has_file['image']:
                            rationale = " + image, no log"
                            reply.set_message('img_only')
                            print("*******************************    image no log") # FIXME bjf
                        elif not self.has_file['lspci']:
                            rationale = " + needs lspci"
                            reply.set_message('also_lspci_vvnn')
                            print("*******************************    needs lspci") # FIXME bjf
                        else:
                            rationale = " + some files attached"
                            reply.set_message('some')
                            print("*******************************    some") # FIXME bjf
        
                    if bug.age_last_message() > 120:
                        rationale = " + Old bug over %d days" % bug.age_last_message()
                        reply.set_message('old_bug')
                        self.add_tag(bug, "needs-verification")
                        print("*******************************    needs-verification") # FIXME bjf

                    if (rationale and self.opt_verbose):
                        print("  --> rationale: %s" % rationale)
        
                    bugline = "%s  %s  %s" % (bug.id, bugtask.status, str(bug.title))

                    # If we've already formulated a reply to this bug, go ahead and
                    # commit that reply, updating the status as well.
                    #
                    if reply.get_message():
                        updated_count += 1
                        total_count   += 1
        
                        if self.opt_dryrun:
                            continue
        
                        text = ""
                        try:
                            text = reply.make_reply()
                            bug.append_comment(text)
                            print bugline
                            self.set_status("Incomplete", bug, bugtask)
                            print("  ---> Incomplete")
                        except:
                            print("  *** Error: Failed to update bug %d\n             Message was: %s\n            New status: %s" % (bug.id, text, new_status))
                            raise
        
                    elif self.has_file['backtrace'] and self.has_file['xorglog'] and self.has_file['lspci']:
                        updated_count += 1
                        total_count   += 1
                        print bugline
                        self.set_status("Confirmed", bug, bugtask)
                        print("  --> Confirmed (has backtrace and log)")
        
                    elif self.has_file['gpudump'] and self.has_file['xorglog'] and self.has_file['lspci']:
                        updated_count += 1
                        total_count   += 1
                        print bugline
                        self.set_status("Confirmed", bug, bugtask)
                        print("  --> Confirmed (has gpudump and log)")
        
                    elif self.has_file['lspci'] and self.has_file['xorglog']:
                        updated_count += 1
                        total_count   += 1
                        print bugline
                        self.set_status("Confirmed", bug, bugtask)
                        print("  --> Confirmed (has log and pci id info)")
        
                    # TODO:  Need to test this in a dryrun first
                    elif 1 == 2:
                        self.action_handler(bug, bugtask)

                    else:
                        if self.opt_verbose:
                            print bugline
                            print("  --> Has all required file attachments.")

                if self.opt_verbose:
                    print("%d bugs updated for %s" % (updated_count, source_pkg))
        
        except:
            # TODO: lazr.restfulclient.errors.HTTPError: HTTP Error 500: Internal Server Error
            if updated_count > 0:
                print("%d bugs updated" % (updated_count))
            raise
        
        if self.opt_verbose:
            print("%d bugs updated" % (total_count))
        
if __name__ == '__main__':
    try:
        app = ProcessNewBugsApp()
        app.main()

    except KeyboardInterrupt:
        # If the user presses <ctrl-C> then stop cleanly
        pass

# vi:set ts=4 sw=4 expandtab:
