#!/usr/bin/env python

import cgi
import gzip
import os
import pdb
import re
import shutil
import sys
import tarfile
import tempfile

from ConfigParser import ConfigParser
from StringIO import StringIO
from fnmatch import fnmatch
from getpass import getpass
from locale import getpreferredencoding
from optparse import OptionParser
from urlparse import urlparse
from zipfile import ZipFile

import pycurl
import gnomekeyring as gkeyring

from arsenal.bugzilla_adapter import *

from lpltk import LaunchpadService
from lpltk.LaunchpadService import LaunchpadServiceError
from lpltk.attachments import *
from lpltk.bug import Bug

class Info(object):
    attrs = ()
    def __init__(self, **kwargs):
        for attr in self.attrs:
            setattr(self, attr, kwargs.get(attr, None))

    def __repr__(self):
        return "<%s>" % u" ".join(["%s: %s" % (attr, getattr(self, attr))
                                   for attr in self.attrs])

    def __str__(self):
        return u"\n".join(["%s: %s" % (attr, getattr(self, attr))
                           for attr in self.attrs])

class LaunchpadInfo(Info):
    attrs = ('url',
             'parsedurl',
             'api',
             'bugid',
             'service')

class BugzillaInfo(Info):
    attrs = ('url',
             'parsedurl',
             'username',
             'password',
             'bugid',
             'attachurl')

class FlagsAndModifiers(type):
    def __init__(cls, name, bases, dct):
        super(FlagsAndModifiers, cls).__init__(name, bases, dict)

        for index, flagname in enumerate(cls.flagnames):
            setattr(cls, flagname, 1 << index)

        for modifier in cls.modifiers:
            setattr(cls, modifier, None)

        def parse_flags(self, flags):
            for flagname in self.flagnames:
                flag = getattr(self, flagname)
                if not flag & flags:
                    setattr(self, flagname, 0)

        def parse_modifiers(self, modifiers):
            for modifier, value in modifiers.iteritems():
                setattr(self, modifier, value)

        setattr(cls, 'parse_flags', parse_flags)
        setattr(cls, 'parse_modifiers', parse_modifiers)

class SendAttachments(object):
    __metaclass__ = FlagsAndModifiers

    flagnames = ('DEBUG',
                 'OWNERONLY',
                 'GZIP',
                 'EXTRACT',
                 'FORCEMIME')

    modifiers = ('bzuser',
                 'bzpassword',
                 'exclude',
                 'xlimit',
                 'maxsize')

    def __init__(self, lpurl, bzurl, flags, modifiers):
        self.tmpdir = None
        self.parse_flags(flags)
        self.parse_modifiers(modifiers)

        self.launchpad = LaunchpadInfo(url=lpurl)
        self.bugzilla = BugzillaInfo(url=bzurl,
                                     user=self.bzuser,
                                     password=self.bzpassword)
        self.parse_bugids()
        self.parse_urls()

        self.print_debug(self.launchpad, self.bugzilla)

    def print_debug(self, *args):
        if self.DEBUG:
            for arg in args:
                print arg
                print

    def parse_bugids(self):
        pattern = re.compile(r'^.*[^\d](\d+)$')
        self.launchpad.bugid = pattern.search(self.launchpad.url).groups()[0]
        self.bugzilla.bugid = pattern.search(self.bugzilla.url).groups()[0]

    def parse_urls(self):
        self.launchpad.parsedurl = urlparse(self.launchpad.url)
        pattern = re.compile(r'^.*?\.(.*)\.(.*)$')
        matches = pattern.search(self.launchpad.parsedurl.netloc).groups()
        self.launchpad.api = (
            self.launchpad.parsedurl.scheme + '://api.' +
            matches[0] + '.' + matches[1])
        # self.launchpad.api = 'edge'

        self.bugzilla.parsedurl = urlparse(self.bugzilla.url)
        pattern = re.compile(r'^(.*)show_bug\.cgi\?id=(\d+)$')
        self.bugzilla.attachurl = pattern.sub(
            r'\1attachment.cgi?bugid=\2&action=enter', self.bugzilla.url)

    def get_user_from_config(self):
        path = os.path.expanduser('~/.config/arsenal')

        if not os.path.isdir(path):
            os.makedirs(path)
        fd = open(os.path.join(path, 'bugzilla.ini'), 'a+')

        config = ConfigParser()
        config.readfp(fd)

        section = self.bugzilla.parsedurl.netloc
        if not config.has_section(section):
            config.add_section(section)

        if not config.has_option(section, 'user'):
            config.set(section, 'user',
                       raw_input('\tEnter Bugzilla username: '))
            config.write(fd)

        fd.close()

        return config.get(section, 'user')

    def get_password_from_keyring(self):
        keyring = 'arsenal'

        if not 'arsenal' in gkeyring.list_keyring_names_sync():
            print "\tCreating new keyring for Arsenal"
            gkeyring.create_sync(keyring, getpass("\tKeyring Password: "))

        info = gkeyring.get_info_sync(keyring)
        if info.get_is_locked():
            print "\tUnlocking keyring for Arsenal"
            gkeyring.unlock_sync(keyring, getpass("\tKeyring Password: "))

        secret = None
        for id in gkeyring.list_item_ids_sync(keyring):
            item = gkeyring.item_get_info_sync(keyring, id)
            if self.bugzilla.username == item.get_display_name():
                attrs = gkeyring.item_get_attributes_sync(keyring, id)
                if attrs['netloc'] == self.bugzilla.parsedurl.netloc:
                    secret = item.get_secret()

        if not secret:
            attrs = {'netloc' : self.bugzilla.parsedurl.netloc}
            id = gkeyring.item_create_sync(keyring,
                                  gkeyring.ITEM_GENERIC_SECRET,
                                  self.bugzilla.username,
                                  attrs,
                                  getpass("\tEnter Bugzilla password: "),
                                  True)
            item = gkeyring.item_get_info_sync(keyring, id)
            secret = item.get_secret()

        return secret

    def login_launchpad(self):
        print "Logging into Launchpad"
        self.launchpad.service = LaunchpadService(config={
                'launchpad_client_name': 'arsenal-send-attachments',
                'launchpad_services_root': self.launchpad.api,
                })
        print ("\t[Success: Logged in as %s]" %
               self.launchpad.service.launchpad.me.display_name)

    def login_bugzilla(self, bzadapter):
        if not self.bugzilla.username:
            self.bugzilla.username = self.get_user_from_config()

        if not self.bugzilla.password:
            self.bugzilla.password = self.get_password_from_keyring()

        if self.bugzilla.username and self.bugzilla.password:
            return bzadapter.login(self.bugzilla.attachurl,
                                        self.bugzilla.username,
                                        self.bugzilla.password)
        else:
            raise BugzillaLoginError("Invalid username/password")

    def run(self):
        self.tmpdir = tempfile.mkdtemp()

        try:
            self.login_launchpad()
        except LaunchpadServiceError as e:
            print "\t[Error: %s]" % e.msg
            return

        bzadapter = BugzillaAdapter(self.tmpdir)
        if bzadapter.is_login_required(self.bugzilla.attachurl):
            try:
                print "Logging into Bugzilla"
                username = self.login_bugzilla(bzadapter)
                if username:
                    print "\t[Success: Logged in as %s]" % username
            except BugzillaLoginError as e:
                print "\t[Error: %s]" % e
                return

        bug = Bug(self.launchpad.service.launchpad, self.launchpad.bugid)
        attachments = Attachments(bug)

        if self.OWNERONLY:
            attachments.add_filter(filter_owned_by_person, [bug.owner,])

        if self.maxsize:
            attachments.add_filter(filter_size_between, [0, self.maxsize])

        if self.exclude:
            attachments.add_filter(filter_filename_matches_globs,
                                   [self.exclude,])

        if self.EXTRACT:
            attachments.extract_archives(self.tmpdir, self.xlimit)

        if self.GZIP:
            attachments.try_gzip(self.tmpdir)

        if self.FORCEMIME:
            attachments.force_mimetype()

        attachments.download_in_dir(self.tmpdir)

        for attachment in attachments:
            print "Uploading attachment <%s>" % attachment.title
            try:
                bzadapter.upload_attachment(attachment,
                                            self.bugzilla.attachurl,
                                            self.bugzilla.bugid)
                print "\t[Success]"
            except BugzillaUploadError as e:
                print "\t[Error: %s]" % e

        shutil.rmtree(self.tmpdir)

if __name__ == "__main__":
    usage = '''
    %prog [OPTIONS] LP_URL BUGZILLA_URL
    '''
    parser = OptionParser(usage=usage)
    parser.add_option('-o', '--owner-only',
        action='store_true', dest='OWNERONLY',
        default=False,
        help='Process only the attachments uploaded by '
             'original bug reporter')
    parser.add_option('-x', '--extract',
        action='store_true', dest='EXTRACT',
        default=False,
        help='Extract tar archives')
    parser.add_option('-g', '--gzip',
        action='store_true', dest='GZIP',
        default=False,
        help='Gzip files which exceed maximum attachment '
             'size')
    parser.add_option('-f', '--force-mime',
        action='store_true', dest='FORCEMIME',
        default=False,
        help='Extract tar archives')
    parser.add_option('-d', '--debug',
        action='store_true', dest='DEBUG',
        default=False,
        help='Enable debugging output')
    parser.add_option('-u', '--user',
        action='store', type='string', dest='user',
        help='Bugzilla username')
    parser.add_option('-p', '--password',
        action='store', type='string', dest='password',
        help='Bugzilla password')
    parser.add_option('-e', '--exclude',
        action='store', type='string', dest='exclude',
        help='Wildcard pattern for attachment exclusion')
    parser.add_option('-l', '--extract-limit',
        action='store', dest='xlimit',
        help='Limit on number of archive members for '
             'extraction')
    parser.add_option('-m', '--max-size',
        action='store', dest='maxsize',
        help='Maximum filesize for processing attachments')
    (options, args) = parser.parse_args()

    if len(args) < 2:
        parser.print_help()
        sys.exit(1)

    flags = 0
    for flagname in SendAttachments.flagnames:
        flags |= getattr(options, flagname) and \
                 getattr(SendAttachments, flagname)

    if options.xlimit:
        try:
            options.xlimit = int(options.xlimit)
        except ValueError:
            options.xlimit = None

    if options.maxsize:
        try:
            options.maxsize = int(options.maxsize)
        except ValueError:
            try:
                prefixes = {'k': 1024, 'm': 1048576, 'g': 1073741824}
                prefix = options.maxsize[-1:].lower()
                options.maxsize = int(options.maxsize[:-1])
                if prefixes.has_key(prefix):
                    options.maxsize = options.maxsize * prefixes[prefix]
                else:
                    raise ValueError
            except ValueError:
                options.maxsize = None

    modifiers = {
        'bzuser' : options.user,
        'bzpassword' : options.password,
        'exclude' : options.exclude,
        'xlimit' : options.xlimit,
        'maxsize' : options.maxsize,
        }

    app = SendAttachments(lpurl=args[0],
        bzurl=args[1],
        flags=flags,
        modifiers=modifiers)
    app.run()

