#!/usr/bin/env python

# Script to assist in forwarding bugs upstream

import os
import re
import sys
import cgi
import tempfile
import shutil
import urllib

from StringIO import StringIO
from locale import getpreferredencoding
from optparse import OptionParser
from urlparse import urlparse

from lpltk import LaunchpadService
from lpltk.LaunchpadService import LaunchpadServiceError
from lpltk.debug import *
from arsenal.arsenal_lib import *
from arsenal.bugzilla_adapter import BugzillaAdapter
from template import Template
from template.util import TemplateException, Literal

# Debugging info
import cgitb
cgitb.enable()
os.environ['LPLTK_DEBUG'] = "1"
os.environ['LPLTOOLKIT_DEBUG'] = "1"

TEMPLATE_PATH    = "/var/lib/arsenal"
SHOW_ATTACHMENTS = 0
DEBUG=1

print "Content-type: text/html\n\n"
dbg("\n\n")

def cleanup_description(desc):
    cleanup_regexes = [
        {'re':'Binary package hint: .*\n\n'},
        {'re':'ProblemType: Bug\n'},
        {'re':'fglrx: Not loaded\n'},
        {'re':'\s*distro:\s*Ubuntu\n'},
        {'re':'\s*SHELL=/bin/bash\n'},
    ]

    for regex in cleanup_regexes:
        regex['rc'] = re.compile(regex['re'], re.IGNORECASE)
        desc = regex['rc'].sub('', desc)

    return desc

def determine_platform(arsenal_bug):
    description_regexes = [
        # x86
        {'platform': 'x86 (IA32)',     're':'Architecture: i386'},
        {'platform': 'x86 (IA32)',     're':'PackageArchitecture: i386'},
        {'platform': 'x86 (IA32)',     're':'Uname: Linux .*generic i686'},
        {'platform': 'x86 (IA32)',     're':'Linux .*-generic .* SMP .* UTC \d+ i686 GNU/Linux'},

        {'platform': 'x86-64 (AMD64)', 're':'Architecture: amd64'},
        {'platform': 'x86-64 (AMD64)',     're':'Uname: Linux .*generic x86_64'},
        {'platform': 'x86-64 (AMD64)', 're':'x86_64'},
    ]
    for regex in description_regexes:
        regex['rc'] = re.compile(regex['re'], re.IGNORECASE)
        if regex['rc'].search(arsenal_bug.description):
            return regex['platform']

    return 'All'

def get_bug_tracker(source_package):
    """ Looks up the bug tracker for the given source package """
    bug_tracker = {
        'tracker':        None,
        'bugzilla_post':  None,
        'product':        None,
        'component':      None,
        }

    product = source_package.upstream_product
    if not product:
        print "<p>Error: Upstream project not known for %s</p>" %(source_package.name)
        return None

    tracker = product.bug_tracker
    if not tracker:
        # Try getting it from the project_group
        project = product.project_group
        if not project:
            print "<p>The bug tracker is not known for project %s</p>" %(product.name)
            print "<p>You can <a href='https://edge.launchpad.net/%s/+configure-bugtracker'>configure %s</a> to specify the bugtracker</p>" %(
                source_package.name, source_package.name)
            return None
        tracker = project.bug_tracker

    if tracker.bug_tracker_type == "Bugzilla":
        bug_tracker['bugzilla_post'] = "/".join([tracker.base_url,"enter_bug.cgi"])
    elif tracker.bug_tracker_type:
        print "<p>Error: This upstream tool only works with Bugzilla"
        print " but %s uses %s as its bug tracker</p>" %(
            source_package.name, tracker.bug_tracker_type)
        print "<p>You are welcome to <a href='https://code.launchpad.net/arsenal/'>add support for this tracker to Arsenal</a>"
        return None
    else:
        print "<p>Error: No bug tracker has been defined for %s</p>" %(
            source_package.name)

    bugzillas = {
        "xserver-xorg-video-ati":{'product':'xorg','component':'Driver/Radeon'},
        "xserver-xorg-video-intel":{'product':'xorg','component':'Driver/intel'},
        "xserver-xorg-video-nouveau":{'product':'xorg','component':'Driver/nouveau'},
        "xserver-xorg-input-evdev":{'product':'xorg','component':'Input/evdev'},
        "xserver-xorg-input-wacom":{'product':'xorg','component':'Input/wacom'},
        "xorg-server":{'product':'xorg','component':'Server/general'},
        "gnome-terminal":{'product':'gnome-terminal','component':'general'},
        "xchat-gnome":{'product':'xchat-gnome','component':'general'},
        "zenity":{'product':'zenity','component':'general'},
        }

    if source_package.name in bugzillas:
        bug_tracker['product'] = bugzillas[source_package.name]['product']
        bug_tracker['component'] = bugzillas[source_package.name]['component']
    else:
        # Make a broad guess that product is just the source package name
        # XXX: Is there a way to validate this in the bugzilla instance?
        bug_tracker['product'] = source_package.name
        bug_tracker['component'] = 'general'

    return bug_tracker

def arsenal_to_bugzilla(arsenal_bug, source_package_name, bug_tracker):
    #lpbug = arsenal_bug.bug.lpbug
    lpbug = arsenal_bug.bug

    original_reporter_name = None
    original_reporter_karma = 0
    if arsenal_bug.owner:
        original_reporter_name = arsenal_bug.owner.display_name.encode('utf-8')
        original_reporter_karma = arsenal_bug.owner.karma

    bug = {
        'bugzilla_post': bug_tracker['bugzilla_post'],

        'id': arsenal_bug.id,
        'short_desc': arsenal_bug.title,
        'bug_severity': 'normal',
        'priority': 'medium',
        'bug_status': 'NEW',
        'original_reporter_name': original_reporter_name,
        'original_reporter_karma': original_reporter_karma,
        'cc': '',

        'product': bug_tracker['product'],
        'token': '',
        'version': '7.5', # TODO
        'source_package': source_package_name,
        'bugzilla_component': bug_tracker['component'],
        'op_sys': 'Linux (All)',
        'rep_platform': determine_platform(arsenal_bug),

        'users_affected': lpbug.users_affected_count,
        'users_unaffected': lpbug.users_unaffected_count,
        'users_total': lpbug.users_affected_count + lpbug.users_unaffected_count,
        'age': arsenal_bug.age(),
        'age_last_message': arsenal_bug.age_last_message(),
        'age_last_updated': arsenal_bug.age_last_updated(),
        'tags': lpbug.tags,

        'problem_statement': arsenal_bug.title,
        'original_description': cleanup_description(arsenal_bug.description),
        'alias': '',
        'keywords': '',
        'attachments': []
        }

    if not lpbug.owner.hide_email_addresses:
        # Look up if reporter's email is registered with bugzilla
        bzurl = "http://bugzilla.freedesktop.org"

        emails = []
        for email_obj in lpbug.owner.confirmed_email_addresses:
            emails.append(email_obj.email)

        bz = BugzillaAdapter("https://bugs.freedesktop.org", arsenal_bug.id)
        email = bz.lookup_registered_email(emails)
        if email:
            bug['cc'] = email
    
    bug['url'] = "http://bugs.launchpad.net/ubuntu/+source/%s/+bug/%d" %(source_package_name, arsenal_bug.id)

    # Get bugtask
    # TODO: Could this be done in a smarter fashion?
    bugtask = None
    for task in lpbug.bug_tasks:
        if not task:
            continue
        elif task.bug_target_display_name != "%s (Ubuntu)" %(source_package_name):
            continue
        else:
            dbg("Found target_name " + task.bug_target_display_name)
            bugtask = task
            exit
    if not bugtask:
        dbg("Could not identify which bugtask to upstream for bug %d" %(arsenal_bug.id))
        return

    # TODO:  Determine severity based off symptom tags

    # Determine if regression based on tags
    if 'regression' in bugtask.bug.tags:
        bug['keywords'] += ' regression'

    # Set bugzilla priority/severity based on launchpad's states
    if (bugtask.importance == 'Critical'):
        dbg("Critical importance")
        bug['priority'] = 'highest'
        bug['bug_severity'] = 'critical'
    elif (bugtask.importance == 'High'):
        dbg("High importance")
        bug['priority'] = 'high'
        bug['bug_severity'] = 'major'
    elif (bugtask.importance == 'Medium'):
        dbg("Medium importance")
        bug['priority'] = 'high'
        bug['bug_severity'] = 'normal'
    elif (bugtask.importance == 'Low'):
        dbg("Low importance")
        bug['priority'] = 'medium'
        bug['bug_severity'] = 'minor'
    elif (bugtask.importance == 'Wishlist'):
        dbg("Wishlist importance")
        bug['priority'] = 'medium'
        bug['bug_severity'] = 'enhancement'
    else:
        dbg("Undefined importance, just using defaults")

    # Get attachments
    # TODO: This is too slow, but maybe it could be improved?
    if SHOW_ATTACHMENTS:
        omit_attachments = [ 'BootDmesg.txt', 'Dependencies.txt', 'Lsusb.txt', 'PciDisplay.txt',
                             'ProcCpuinfo.txt', 'ProcInterrupts.txt', 'ProcModules.txt',
                             'UdevDb.txt', 'UdevLog.txt', 'setxkbmap.txt', 'xkbcomp.txt' ]
        for a in arsenal_bug.attachments:
            deb("Processing an attachment")
            try:
                hosted_file = a.data
                hosted_file_buffer = hosted_file.open()
                hosted_file_name = hosted_file_buffer.filename

                # TODO: Use the json
                attachment = {
                    'file_name': hosted_file_name,
                    'title': a.title,
                    'subject': a.message.subject,
                    'owner': a.message.owner.display_name,
                    'created': a.message.date_created,
                    'ispatch': a.type,
                    'include': 1,
                    'url': "%s" %(a.self),
                    'http_etag': "%s" %(a.http_etag),
                    'contenttypemethod': 'manual',
                    'contenttypeentry': 'text/plain'
                    }
#                if a.message.owner.display_name != arsenal_bug.owner.display_name:
#                    attachment['include'] = 0
                if a.title in omit_attachments:
                    attachment['include'] = 0

                bug['attachments'].append(attachment)
            except:
                print "<p>Error:  Could not read attachment</p>"
                raise

    return bug

# TODO: Replace the following stuff with genshi
TEMPLATE_CONTENT = "templates/upstream_form.tt2"
TEMPLATE_MAIN    = "templates/main.tt2"
def upstream_form(bug):
    tt2 = Template({'INCLUDE_PATH': TEMPLATE_PATH})
    text = ''

    if not bug:
        return None

    try:
        dbg("Processing content template")
        content = tt2.process(TEMPLATE_CONTENT, bug )
        dbg("Processing main template")
        text = tt2.process(TEMPLATE_MAIN, {
            'title': "Forward LP: #%d to FDO" % (bug['id']),
            'content': content
            } )
        dbg("Finished template processing")
    except TemplateException, e:
        dbg("ERROR: %s" % (e))

    return text

def bug_form():
    text  = "<form action='send_upstream.cgi' method='GET'>"
    text += "<p>LP Bug# <input type='text' name='bug' size='6' />"
    text += "<input type='submit' name='Action' value='Go'></p>"
    text += "</form>"
    return text

def update_launchpad(arsenal_bug, fdo_bug):
    dbg("Unimplemented")
    # TODO:  Add bugtask with watch on fdo_bug

    if fdo_bug['cc']:
        message = """Hi $name,
        I\'ve forwarded your bug upstream to $bz_url and subscribed you to it.  Please follow up there in case upstream needs more information or wishes you to test something.  Thanks ahead of time!
        """
    else:
        message = """Hi $name,
        I\'ve forwarded your bug upstream to $bz_url - please subscribe to this bug in case they need more information or wish you to test something.  Thanks ahead of time!
        """
    arsenal_bug.append_comment(message)

def main():
    tmphome = tempfile.mkdtemp(prefix='arsenal')
    os.environ['HOME'] = tmphome
    try:
        lp          = LaunchpadService(config={'read_only':True})
        d           = lp.load_project("ubuntu")
    except:
        print "<p>Error:  Failed to access Launchpad.</p>\n"
        shutil.rmtree(tmphome)
        raise

    # Process form contents
    form = cgi.FieldStorage()
    if (form.has_key("bug")):
        lpbug = lp.launchpad.bugs[form["bug"].value]
        arsenal_bug = ArsenalBug(lpbug, lp.launchpad)

        # TODO: Could this be done more cleverly?
        #       Perhaps use the fact that we already loaded the project above?
        sources = arsenal_bug.distro_task_sources("Ubuntu")

        if len(sources) > 1:
            # TODO: Give the user a choice to select from
            print "<p>This bug report has too many source packages</p>"
            return 1
        elif len(sources) < 1:
            print "<p>This bug report is not open against any source packages that can be upstreamed</p>"
            return 1
        source_package_name = sources[0]

        source_package = d.getSourcePackage(name=source_package_name)
        if not source_package:
            print "<p>Error: Source Package could not be found for %s</p>" %(source_package_name)
            return 1

        bug_tracker = get_bug_tracker(source_package)
        if not bug_tracker:
            return 1

        bug = arsenal_to_bugzilla(arsenal_bug, source_package_name, bug_tracker)
        print upstream_form(bug)

    elif (form.has_key("TODO")):
        # Send bug to the upstream bugzilla (in a new window if possible)
        # TODO:  Hook in kamran's attachment sender here
        # TODO:  Add python-bzutils support
        bug = "bugzilla bug info here"
        update_launchpad(bug)
    else:
        print "<br/>"
        print "<blockquote>"
        print "<p>Upstream bug from Launchpad</p>"
        print bug_form()
        print """
<br/>
<br/>
<h3>Notes</h3>
<p>
In theory this tool should be able to forward any bug report, however
there are some limitations.</p>

<ol>
<li> If a bug is filed against more than one source package, the tool
     can't figure out what to file against.  (Ugly workaround: Close all
     bug tasks except the one to send upstream.)</li>
<li> Launchpad doesn't map source packages to upstream components.  We
     just guess the component to be 'general' and the product to be the
     source package name, which seems to be how GNOME does it.</li>
<li> I've special cased a variety of X.org packages, since Freedesktop
     has a more complex package name mapping.  Most common X.org packages
     should work.</li>
</ol>

</blockquote>
"""

    shutil.rmtree(tmphome)
    return 0

### Main program ###
if __name__ == '__main__':
    sys.exit(main())
